@@ -1,6 +1,6 @@
 /* mpf_ui_sub -- Subtract a float from an unsigned long int.
 
-Copyright 1993, 1994, 1995, 1996, 2001 Free Software Foundation, Inc.
+Copyright 1993, 1994, 1995, 1996, 2001, 2002 Free Software Foundation, Inc.
 
 This file is part of the GNU MP Library.
 
@@ -210,9 +210,9 @@ mpf_ui_sub (mpf_ptr r, unsigned long int u, mpf_srcptr v)
 		      /* uuuu     */
 		      mp_size_t size, i;
 		      size = usize - vsize;
-		      tp[0] = -up[0];
+		      tp[0] = -up[0] & GMP_NUMB_MASK;
 		      for (i = 1; i < size; i++)
-			tp[i] = ~up[i];
+			tp[i] = ~up[i] & GMP_NUMB_MASK;
 		      mpn_sub_n (tp + size, vp, up + size, vsize);
 		      mpn_sub_1 (tp + size, tp + size, vsize, (mp_limb_t) 1);
 		      negate ^= 1;
@@ -229,9 +229,9 @@ mpf_ui_sub (mpf_ptr r, unsigned long int u, mpf_srcptr v)
 		    {
 		      mp_size_t size, i;
 		      size = vsize - usize;
-		      tp[0] = -vp[0];
+		      tp[0] = -vp[0] & GMP_NUMB_MASK;
 		      for (i = 1; i < size; i++)
-			tp[i] = ~vp[i];
+			tp[i] = ~vp[i] & GMP_NUMB_MASK;
 		      mpn_sub_n (tp + size, up, vp + size, usize);
 		      mpn_sub_1 (tp + size, tp + size, usize, (mp_limb_t) 1);
 		      rsize = vsize;
@@ -287,9 +287,9 @@ mpf_ui_sub (mpf_ptr r, unsigned long int u, mpf_srcptr v)
 		  /*   vvvvv  */
 		  mp_size_t size, i;
 		  size = vsize + ediff - usize;
-		  tp[0] = -vp[0];
+		  tp[0] = -vp[0] & GMP_NUMB_MASK;
 		  for (i = 1; i < size; i++)
-		    tp[i] = ~vp[i];
+		    tp[i] = ~vp[i] & GMP_NUMB_MASK;
 		  mpn_sub (tp + size, up, usize, vp + size, usize - ediff);
 		  mpn_sub_1 (tp + size, tp + size, usize, (mp_limb_t) 1);
 		  rsize = vsize + ediff;
@@ -302,11 +302,11 @@ mpf_ui_sub (mpf_ptr r, unsigned long int u, mpf_srcptr v)
 	  /*      vv  */
 	  mp_size_t size, i;
 	  size = vsize + ediff - usize;
-	  tp[0] = -vp[0];
+	  tp[0] = -vp[0] & GMP_NUMB_MASK;
 	  for (i = 1; i < vsize; i++)
-	    tp[i] = ~vp[i];
+	    tp[i] = ~vp[i] & GMP_NUMB_MASK;
 	  for (i = vsize; i < size; i++)
-	    tp[i] = ~(mp_limb_t) 0;
+	    tp[i] = GMP_NUMB_MAX;
 	  mpn_sub_1 (tp + size, up, usize, (mp_limb_t) 1);
 	  rsize = size + usize;
 	}
