@@ -1,21 +1,21 @@
 /* gmp_randseed (state, seed) -- Set initial seed SEED in random state
    STATE.
 
-Copyright (C) 2000  Free Software Foundation, Inc.
+Copyright 2000, 2004 Free Software Foundation, Inc.
 
 This file is part of the GNU MP Library.
 
 The GNU MP Library is free software; you can redistribute it and/or modify
-it under the terms of the GNU Library General Public License as published by
-the Free Software Foundation; either version 2 of the License, or (at your
+it under the terms of the GNU Lesser General Public License as published by
+the Free Software Foundation; either version 2.1 of the License, or (at your
 option) any later version.
 
 The GNU MP Library is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
+or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 License for more details.
 
-You should have received a copy of the GNU Library General Public License
+You should have received a copy of the GNU Lesser General Public License
 along with the GNU MP Library; see the file COPYING.LIB.  If not, write to
 the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 MA 02111-1307, USA. */
@@ -24,14 +24,10 @@ MA 02111-1307, USA. */
 #include "gmp-impl.h"
 
 void
-#if __STDC__
 gmp_randseed (gmp_randstate_t rstate,
-	      mpz_t seed)
-#else
-gmp_randseed (rstate, seed)
-     gmp_randstate_t rstate;
-     mpz_t seed;
-#endif
+	      mpz_srcptr seed)
 {
-  mpz_set (rstate->seed, seed);
+  mpz_fdiv_r_2exp (rstate->_mp_seed, seed,
+                   rstate->_mp_algdata._mp_lc->_mp_m2exp);
+
 }
